/*
    SmartHomeGeeks.io Smart Doorbell
    https://smarthomegeeks.io/smart-doorbell
*/

#include <FS.h>
#include <WiFiManager.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoJson.h>
#include <PubSubClient.h>
#include <ArduinoOTA.h>

// Device configuration
const String deviceName = "Doorbell";

const int switch1 = 4;
const int relay1 = 12;

char mqtt_server[40];
char mqtt_port[6] = "1883";
char mqtt_user[96];
char mqtt_pass[96];
char mqtt_client[40] = "doorbell";
bool chime = true;
bool button = false;
unsigned long startMillis = millis();
unsigned long lastMQTTconnect = 0;

WiFiClient espClient;
PubSubClient client(espClient);

//flag for saving data
bool shouldSaveConfig = false;

//callback notifying us of the need to save config
void saveConfigCallback () {
  Serial.println("Should save config");
  shouldSaveConfig = true;
}

void autoDiscovery() {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[512];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["ids"] = mqtt_client;
  deviceData["mf"] = "SmartHomeGeeks.io";
  deviceData["sw"] = "2023.08";
  
  // Switch
  DynamicJsonDocument root(512);
  root["name"] = "Chime";
  root["obj_id"] = String(mqtt_client) + "_chime";
  root["ic"] = "mdi:bell-alert";
  root["uniq_id"] = "switch." + String(mqtt_client);
  root["stat_t"] = String(mqtt_client) + "/chimeStatus";
  root["cmd_t"] = String(mqtt_client) + "/chimeCmd";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "config";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/switch/" + String(mqtt_client) + "/config").c_str(), buffer, true);
  root.clear();

  // Button 
  root["name"] = nullptr;
  root["obj_id"] = String(mqtt_client) + "_button";
  root["ic"] = "mdi:bell-ring";
  root["uniq_id"] = "button." + String(mqtt_client);
  root["cmd_t"] = String(mqtt_client) + "/buttonCmd";
  root["avty_t"] = String(mqtt_client) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/button/" + String(mqtt_client) + "/config").c_str(), buffer, true);
  root.clear();

  // Button Sensor
  root["name"] = nullptr;
  root["obj_id"] = String(mqtt_client) + "_button";
  root["ic"] = "mdi:doorbell";
  root["uniq_id"] = "sensor." + String(mqtt_client);
  root["stat_t"] = String(mqtt_client) + "/buttonStatus";
  root["avty_t"] = String(mqtt_client) + "/availability";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/binary_sensor/" + String(mqtt_client) + "/config").c_str(), buffer, true);
  root.clear();

  // RSSI sensor
  root["name"] = "RSSI";
  root["obj_id"] = String(mqtt_client) + "_rssi";
  root["stat_t"] = String(mqtt_client) + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + String(mqtt_client) + "_rssi";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqtt_client) + "/rssi/config").c_str(), buffer, true);
  root.clear();

  // IP
  root["name"] = "IP";
  root["obj_id"] = String(mqtt_client) + "_ip";
  root["stat_t"] = String(mqtt_client) + "/ip";
  root["ic"] = "mdi:ip-network";
  root["uniq_id"] = "sensor." + String(mqtt_client) + "_ip";
  root["avty_t"] = String(mqtt_client) + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + String(mqtt_client) + "/ip/config").c_str(), buffer, true);
  root.clear();

  client.publish((String(mqtt_client) + "/buttonStatus").c_str(), "OFF", true);
  client.publish((String(mqtt_client) + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);

}

void connectmqtt() {
  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqtt_client, mqtt_user, mqtt_pass, (String(mqtt_client) + "/availability").c_str(), 1, true, "offline")) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);
    autoDiscovery();

    client.subscribe((String(mqtt_client) + "/buttonCmd").c_str());
    client.subscribe((String(mqtt_client) + "/chimeCmd").c_str());
    client.publish((String(mqtt_client) + "/availability").c_str(), "online", true);
    client.publish((String(mqtt_client) + "/ip").c_str(),  WiFi.localIP().toString().c_str(), true);

  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }
}

void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);
  Serial.println();

  pinMode(switch1, INPUT_PULLUP);
  pinMode(relay1, OUTPUT);

  //clean FS, for testing
  // SPIFFS.format();

  //read configuration from FS json
  Serial.println("mounting FS...");

  if (SPIFFS.begin()) {
    Serial.println("mounted file system");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);

        DynamicJsonDocument json(1024);
        auto deserializeError = deserializeJson(json, buf.get());
        serializeJson(json, Serial);
        if ( ! deserializeError ) {
          Serial.println("\nparsed json");
          strcpy(mqtt_server, json["mqtt_server"]);
          strcpy(mqtt_port, json["mqtt_port"]);
          strcpy(mqtt_user, json["mqtt_user"]);
          strcpy(mqtt_pass, json["mqtt_pass"]);
          strcpy(mqtt_client, json["mqtt_client"]);
          chime = json["chime"];
          client.publish((String(mqtt_client) + "/chimeStatus").c_str(), (chime ? "ON" : "OFF"), true);
        } else {
          Serial.println("failed to load json config");
        }
        configFile.close();
      }
    }
  } else {
    Serial.println("failed to mount FS");
  }
  //end read

  // The extra parameters to be configured (can be either global or just in the setup)
  // After connecting, parameter.getValue() will get you the configured value
  // id/name placeholder/prompt default length
  WiFiManagerParameter custom_mqtt_server("mqtt_server", "MQTT server", mqtt_server, 40);
  WiFiManagerParameter custom_mqtt_port("mqtt_port", "MQTT port", mqtt_port, 6);
  WiFiManagerParameter custom_mqtt_user("mqtt_user", "MQTT Username", mqtt_user, 64);
  WiFiManagerParameter custom_mqtt_pass("mqtt_pass", "MQTT Password", mqtt_pass, 64);
  WiFiManagerParameter custom_mqtt_client("mqtt_client", "MQTT Topic", mqtt_client, 40);

  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
  WiFiManager wifiManager;

  // Set WiFi Hostname
  WiFi.hostname(mqtt_client);

  //set config save notify callback
  wifiManager.setSaveConfigCallback(saveConfigCallback);

  //add all your parameters here
  wifiManager.addParameter(&custom_mqtt_server);
  wifiManager.addParameter(&custom_mqtt_port);
  wifiManager.addParameter(&custom_mqtt_user);
  wifiManager.addParameter(&custom_mqtt_pass);
  wifiManager.addParameter(&custom_mqtt_client);

  // reset settings - for testing
  // wifiManager.resetSettings();

  // configure a portal timeout and restart
  wifiManager.setConfigPortalTimeout(300);

  if (!wifiManager.autoConnect("SHGio_Doorbell")) {
    Serial.println("failed to connect and hit timeout");
    delay(5000);
    ESP.restart();
  }

  //if you get here you have connected to the WiFi
  Serial.println("connected...yeey :)");

  //read updated parameters
  strcpy(mqtt_server, custom_mqtt_server.getValue());
  strcpy(mqtt_port, custom_mqtt_port.getValue());
  strcpy(mqtt_user, custom_mqtt_user.getValue());
  strcpy(mqtt_pass, custom_mqtt_pass.getValue());
  strcpy(mqtt_client, custom_mqtt_client.getValue());
  Serial.println("The values in the file are: ");
  Serial.println("\tmqtt_server : " + String(mqtt_server));
  Serial.println("\tmqtt_port : " + String(mqtt_port));
  Serial.println("\tmqtt_user : " + String(mqtt_user));
  Serial.println("\tmqtt_pass : " + String(mqtt_pass));
  Serial.println("\tmqtt_client : " + String(mqtt_client));
  Serial.println("\tchime : " + String(chime));

  if (shouldSaveConfig) saveConfig();

  Serial.println("local ip");
  Serial.println(WiFi.localIP());

  // Configure OTA
  ArduinoOTA.setHostname(mqtt_client);
  ArduinoOTA.begin(); 

  // Connect MQTT
  client.setServer(mqtt_server, atoi(mqtt_port));
  client.setCallback(callback);
}

void saveConfig() {
  //save the custom parameters to FS
    Serial.println("saving config");
    DynamicJsonDocument json(1024);
    json["mqtt_server"] = mqtt_server;
    json["mqtt_port"] = mqtt_port;
    json["mqtt_user"] = mqtt_user;
    json["mqtt_pass"] = mqtt_pass;
    json["mqtt_client"] = mqtt_client;
    json["chime"] = chime;

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("failed to open config file for writing");
    }

    serializeJson(json, Serial);
    serializeJson(json, configFile);
    configFile.close();
    //end save
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Topic arrived: ");
  Serial.println(String(topic));
  
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  if (String(topic) == String(mqtt_client) + "/chimeCmd") {
    chime = (!strncmp((char*)payload, "ON", length) ? true : false);
    client.publish((String(mqtt_client) + "/chimeStatus").c_str(), (!strncmp((char*)payload, "ON", length) ? "ON" : "OFF"), true);
    Serial.print("Changed chime value to: ");
    Serial.println(String(chime));
    saveConfig();
  } else if (String(topic) == String(mqtt_client) + "/buttonCmd" && !strncmp((char*)payload, "PRESS", length))  {
    chimeOn();
  }

  client.publish((String(mqtt_client) + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
}

void chimeOn() {
  button = true;
  startMillis = millis();
  if (chime) digitalWrite(relay1, HIGH);
  client.publish((String(mqtt_client) + "/buttonStatus").c_str(), "ON", true);
}

void chimeOff() {
  button = false;
  client.publish((String(mqtt_client) + "/buttonStatus").c_str(), "OFF", true);
}

void loop() {
  if (!client.connected()) connectmqtt();

  // doorbell button pressed
  if (!button && digitalRead(switch1) == LOW) chimeOn();

  else if (button && (millis() - startMillis) > 600) {
    // stop the chime when button is pressed over long
    digitalWrite(relay1, LOW);

    // doorbell buton released
    if (digitalRead(switch1) == HIGH) chimeOff();
  }

  ArduinoOTA.handle();

  client.loop();

  delay(200);

}